import os
import sys
import logging
import subprocess
import signal

from kocher_tools.misc import confirmExecutable

def assignOutput (out_path, discard_empty_output, barcode_type):

	# Create list to hold commands
	output_list = []

	# Define the basic filename
	output_filename = os.path.join(out_path, '%%_%s.fastq.gz')

	# Check if the barcode type is i5
	if barcode_type == 'i5':

		# Check if the empty output should not be assigned
		if discard_empty_output:

			# Indicate that no i5 output should be produced
			output_list.extend(['-o', 'n/a']) 

		# Check if the empty output should be assigned
		else:

			# Assign the i5 output
			output_list.extend(['-o', output_filename % 'i5'])

		# Assign the i7 file
		output_list.extend(['-o', output_filename % 'i7'])

	# Check if the barcode type is i7
	elif barcode_type == 'i7':
	
		# Check if the empty output should not be assigned
		if discard_empty_output:

			# Indicate that no i5 output should be produced
			output_list.extend(['-o', 'n/a']) 

		# Check if the empty output should be assigned
		else:

			# Assign the i5 output
			output_list.extend(['-o', output_filename % 'i7'])

	# Assign the R1 and R2 output
	output_list.extend(['-o', output_filename % 'R1'])
	output_list.extend(['-o', output_filename % 'R2'])

	return output_list

def i5ReformatMap (i5_map_filename, reformatted_filename):

	# Open the reformatted file
	reformatted_i5_map = open(reformatted_filename, 'w')

	# Open the i5 map file
	with open(i5_map_filename) as i5_map_file:

		# Loop the i5 map file, line by line
		for i5_map_line in i5_map_file:

			# Split the line into: plate, barcode, locus
			i5_plate, i5_barcode, i5_locus = i5_map_line.split()

			# Write to the reformatted file
			reformatted_i5_map.write('%s_%s\t%s\n' % (i5_plate, i5_locus, i5_barcode))

	# Close the file
	reformatted_i5_map.close()	

def i5BarcodeJob (i5_map_filename, i5_input, i7_input, r1_input, r2_input, out_path, discard_i5):

	# Define the reformatted i5 map filename
	reformatted_i5_map_filename = i5_map_filename + '.reformatted'

	# Reformat the i5 map
	i5ReformatMap(i5_map_filename, reformatted_i5_map_filename)

	# Create the basic input arg list
	multiplex_call_args = ['-B', reformatted_i5_map_filename, i5_input, i7_input, r1_input, r2_input]

	# Add the output args, using the path, if empty files should be kept, and set the barcode type as i5
	multiplex_call_args.extend(assignOutput(out_path, discard_i5, 'i5'))

	# Call fastq-multz with the argus
	callFastqMultx(multiplex_call_args)

	# Remove the reformatted i5 map
	os.remove(reformatted_i5_map_filename)

def i7BarcodeJob (i7_map_filename, i7_input, r1_input, r2_input, out_path, discard_i7):

	# Create the basic input arg list
	multiplex_call_args = ['-B', i7_map_filename, i7_input, r1_input, r2_input]

	# Add the output args, using the path, if empty files should be kept, and set the barcode type as i5
	multiplex_call_args.extend(assignOutput(out_path, discard_i7, 'i7'))

	# Call fastq-multz with the argus
	callFastqMultx(multiplex_call_args)

def checkFastqMultxForErrors (fastq_multx_stderr):

	# Loop the stderr line by line
	for fastq_multx_stderr_lines in fastq_multx_stderr:

		# Ignore barcode file statement
		if 'Using Barcode File:' in fastq_multx_stderr:
			pass

		# Ignore end used file statement
		elif 'End used:' in fastq_multx_stderr:
			pass

		# Ignore the skipped (due to distance) statement
		elif 'Skipped because of distance' in fastq_multx_stderr:
			pass

		# Check if there is an error
		elif 'Error:' in fastq_multx_stderr:

			# Report the error
			raise Exception(fastq_multx_stderr)

		# Check if there are other messages
		else:

			print ('Error? - %s' % fastq_multx_stderr)

def callFastqMultx (fastq_multx_call_args):

	fastq_multx_executable = confirmExecutable('fastq-multx')

	# Check if executable is installed
	if not fastq_multx_executable:
		raise IOError('fastq-multx not found. Please confirm the executable is installed')

	# fastq-multx subprocess call
	fastq_multx_call = subprocess.Popen([fastq_multx_executable] + fastq_multx_call_args, stderr = subprocess.PIPE, stdout = subprocess.PIPE, preexec_fn = lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL))

	# Get stdout and stderr from subprocess
	fastq_multx_stdout, fastq_multx_stderr = fastq_multx_call.communicate()

	# Check if code is running in python 3
	if sys.version_info[0] == 3:
		
		# Convert bytes to string
		fastq_multx_stdout = fastq_multx_stdout.decode()
		fastq_multx_stderr = fastq_multx_stderr.decode()

	# Check the stderr for errors
	checkFastqMultxForErrors(fastq_multx_stderr)


import os
import sys
import subprocess
import gzip

from kocher_tools.misc import confirmExecutable

def gzipCompress (gzip_filename, return_filename = False, overwrite = True):

	# Find the gzip executable
	gzip_executable = confirmExecutable('gzip')

	# Check if executable is installed
	if not gzip_executable:
		raise IOError('gzip not found. Please confirm the executable is installed')

	# Assign the overwrite argument
	overwrite_arg = []

	# Check if gzip shouldnt overwrite
	if overwrite:

		# Add the overwrite arg
		overwrite_arg.append('-f')

	# vsearch subprocess call
	gzip_call = subprocess.Popen([gzip_executable, gzip_filename] + overwrite_arg, stderr = subprocess.PIPE, stdout = subprocess.PIPE)

	# Get stdout and stderr from subprocess
	gzip_stdout, gzip_stderr = gzip_call.communicate()

	# Check if code is running in python 3
	if sys.version_info[0] == 3:
		
		# Convert bytes to string
		gzip_stdout = gzip_stdout.decode()
		gzip_stderr = gzip_stderr.decode()

	# Update when starting logs
	#print(gzip_stdout)
	#print(gzip_stderr)

	# Check if the filename is to be returned
	if return_filename:
		return gzip_filename + '.gz'

def gzipIsEmpty (gzip_filename):

	# Open the gzip file
	with gzip.open(gzip_filename, 'rb') as gzip_file:

		# Read the file
		gzip_data = gzip_file.read(1)

	# Check if the file is empty
	if len(gzip_data) == 0:

		# Return true, if empty
		return True

	# Otherwise, return false
	return False


#!/usr/bin/env python
import os
import sys
import csv
import argparse
import pandas as pd
from collections import defaultdict
from kocher_tools.gff import *

def gffArgs ():

	'''
	Argument parser

	Raises
	------
	IOError
		If the input, or other specified files do not exist
	'''

	def confirmFile ():
		'''Custom action to confirm file exists'''
		class customAction(argparse.Action):
			def __call__(self, parser, args, value, option_string=None):
				if not os.path.isfile(value):
					raise IOError('%s not found' % value)
				setattr(args, self.dest, value)
		return customAction

	gff_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	# Assign I/O arguments
	gff_parser.add_argument('--gff-file', help = "Defines the GFF filename", type = str, action = confirmFile(), required = True)
	gff_parser.add_argument('--chrom-size-file', help = "Defines the chromosome size filename (as tsv)", type = str, action = confirmFile(), required = True)
	gff_parser.add_argument('--out-file', help = "Defines the output filename (as tsv). Defaults to out.tsv", type = str, default = 'out.tsv')
	gff_parser.add_argument('--delete-db', help = "Defines if the database should be deleted when finished", action = 'store_true')

	# Assign the bp values for added features
	gff_parser.add_argument('--promoter-bp', help = "Defines the promoter length", type = int, default = 5000)
	gff_parser.add_argument('--upstream-bp', help = "Defines the upstream length", type = int, default = 10000)
	gff_parser.add_argument('--downstream-bp', help = "Defines the downstream length", type = int, default = 10000)

	'''
	Assign features arguments. Some features appear multiple times due to alternative 
	feature names - i.e. ['five_prime_UTR', 'five_utr']. This will not cause an error
	as features not found within the file will be ignored.
	'''
	feature_set = ['tss_upstream', 'tss_flanks', 'five_prime_UTR', 'five_utr', 'exon', 'CDS', 'first_intron', 'intron', 'three_prime_UTR', 'three_utr', 'intergenic']
	gff_parser.add_argument('--feature-set', help = "Defines the feature set. Feature not it list will be ignored", type = str, nargs = '+', default = feature_set)
	priority_list = ['tss_upstream', 'tss_flanks', 'promoter', 'upstream', 'downstream', 'five_prime_UTR', 'five_utr', 'exon', 'CDS',  'first_intron', 'intron', 'three_prime_UTR', 'three_utr', 'intergenic']
	gff_parser.add_argument('--priority-order', help = "Defines the priority of features", type = str, nargs = '+', default = priority_list)

	# Return arguments
	return gff_parser.parse_args()

# Assign the arguments using argparse
gff_args = gffArgs()

# Assign a set with all the possible features
feature_set = gff_args.feature_set + ['upstream', 'promoter', 'downstream']

# Check if features in the priority are not in the feature set
unique_priority_features = list(set(gff_args.priority_order) - set(feature_set))
if unique_priority_features: raise Exception(f'Feature(s) found in --priority-order not in --feature-set: %s' % ', '.join(unique_priority_features))

# Create the database
gff_db = createDatabase(gff_args.gff_file)

'''
Read in the chromosome sizes file and confirm the sizes are ints
then read in the chromosome positions file and group the positions
'''
chrom_sizes = pd.read_csv(gff_args.chrom_size_file, sep = '\t', header = None, usecols=[0, 1], dtype = {0: str, 1: int})

try:
	# Create defaultdict to store the feature counts
	feature_counts_no_priority = defaultdict(int)
	feature_counts_w_priority = defaultdict(int)

	# Loop the GFF database by each chromosome, use the size to know when to stop iter
	for chrom_name, chrom_size in chrom_sizes.values:
		chrom_feature_counts_w_priority = defaultdict(int)
		chrom_feature_counts_w_priority = chromCounts(gff_db, chrom_name, chrom_size, chrom_feature_counts_w_priority, prioritize = True, **vars(gff_args))
		if sum(chrom_feature_counts_w_priority.values()) != chrom_size: raise Exception(f'Unable to parse, likely contains additional GFF entries: {chrom_name}')
		for _f, _c in chrom_feature_counts_w_priority.items(): feature_counts_w_priority[_f] += _c
		feature_counts_no_priority = chromCounts(gff_db, chrom_name, chrom_size, feature_counts_no_priority, **vars(gff_args))

	# Check if features in the priority list were not found in the GFF
	missing_features = list(set(gff_args.priority_order) - set(list(feature_counts_no_priority)))
	if missing_features: 
		missing_features_str = 'Feature(s) not in gff: %s' % ', '.join(missing_features)
		print(missing_features_str)

	# Save the counts into a dataframe, write the dataframe to a tsv
	output_file = open(gff_args.out_file, 'w')
	output_file.write('#%s\n' % ' '.join(sys.argv))
	if missing_features: output_file.write('#%s\n' % missing_features_str)
	output_dataframe = pd.DataFrame.from_dict({'Total Feature Counts': feature_counts_no_priority, 'Feature Counts w/ Priority': feature_counts_w_priority}, orient='index')
	output_dataframe = output_dataframe.fillna(0)
	output_dataframe = output_dataframe[[_pc for _pc in gff_args.priority_order if _pc in output_dataframe.columns]]
	output_dataframe = output_dataframe.astype(int)
	output_dataframe.to_csv(output_file, sep = '\t')
	output_file.close()

	# Delete the database, if specified
	if gff_args.delete_db: os.remove(f'{gff_args.gff_file}.db')

except:	raise


import os
import json
import logging

import pandas as pd
import numpy as np

from dateutil import parser
from Bio import SeqIO

from kocher_tools.config_file import ConfigDB
from kocher_tools.database import *
from kocher_tools.assignment import readInputFile

def insertCollectionFileUsingConfig (config_file, schema, filepath, uploader, ignore_previously_entered):

	def check_date (data, date = False):
		try:
			date_time = parser.parse(data, fuzzy=True)
			if date: date_time = date_time.date()
			return date_time
		except:
			return ''

	# Open the config and start the SQL session
	config_data = ConfigDB.readConfig(config_file)
	sql_connection = startSessionFromConfig(config_data)

	try:

		# Read in the file as a pandas dataframe and convert labels, if needed
		input_dataframe = readInputFile(filepath)
		colleciton_label_dict = {'Barcode': 'Unique ID',
								 'Date': 'Date Collected',
								 'Time': 'Time Entered',
								 'Has Pollen': 'Has Pollen',
								 'Preservation Method': 'Sample Preservation Method',
								 'Head in PFA?': 'Head Preserved',
								 'From Nest': 'From Nest',
								 'Field_ID': 'Species Guess'}
		input_dataframe = input_dataframe.rename(columns=colleciton_label_dict)

		# Prep the dataframe
		input_dataframe = prepDataFrameUsingConfig(config_data, schema, input_dataframe)
		input_dataframe['collection_file'] = os.path.basename(filepath)
		if uploader: input_dataframe['collected_by'] = ' '.join(uploader)

		# Check for the unique ID column
		if 'unique_id' not in input_dataframe.columns: raise Exception(f'Unable to assign IDs, cannot locate the assignment column')

		# Update the dates
		input_dataframe['date_collected'] = input_dataframe['date_collected'].apply(check_date, date = True)
		input_dataframe['date_collected'] = pd.to_datetime(input_dataframe['date_collected']).dt.date

		# Clean up the data, with dates
		input_dataframe = input_dataframe.replace(np.nan, None)

		# Assign the table
		sql_table_assign = config_data[schema]

		# Ignore - i.e. don't raise an exception - if previously entered data is found
		if ignore_previously_entered:

			# Check for previously entered data, and filter out
			sql_column_assign = config_data.getSQLColumn(sql_table_assign.columns, 'unique_id')
			sql_select = SQLSelect.fromConfig(config_data, sql_connection)
			sql_select.addColumnToSelect(sql_column_assign)
			sql_select.addDictWhere({'unique_id':input_dataframe['unique_id'].values}, include = True, cmp_type = 'IN', dict_type = 'Column')
			sql_select.select()
			select_dataframe = sql_select.toDataFrame()
			input_dataframe = input_dataframe[~input_dataframe['unique_id'].isin(select_dataframe['unique_id'].values)]

			# Check if all the data was removed
			if input_dataframe.empty: 
				logging.warning(f'No new collection samples.')
				sql_connection.close()
				return

		# Insert the dataframe into the database
		sql_table_assign = config_data[schema]
		sql_insert = SQLInsert.fromConfig(config_data, sql_connection)
		sql_insert.addTableToInsert(sql_table_assign)
		sql_insert.addDataFrameValues(input_dataframe)
		sql_insert.insert()

	except:
		sql_connection.rollback()
		raise

	else:
		sql_connection.commit()

def insertBarcodeFilesUsingConfig (config_file, schema, filepaths):

	def append_seq (query_id, seq_index):
		return seq_index[query_id].format('fasta')

	def join_list (value_list):
		try: return ', '.join([value_str for value_str in value_list if value_str != 'BOLD:N/A'])
		except: return ''

	# Open the config and start the SQL session
	config_data = ConfigDB.readConfig(config_file)
	sql_connection = startSessionFromConfig(config_data)

	try:
		# Assign the expected filetypes
		blast_filepath = None
		fasta_filepath = None
		failed_filepath = None

		# Loop the filepaths
		for filepath in filepaths:

			# Assign the blast file
			if filepath.endswith('.out'):
				if blast_filepath: raise Exception(f'BLAST file already assigned ({blast_filepath}), cannot assign: {filepath}')
				blast_filepath = filepath

			# Assign the fasta file
			if filepath.endswith('.fasta') or filepath.endswith('.fa') or filepath.endswith('.fas'):
				if fasta_filepath: raise Exception(f'FASTA file already assigned ({fasta_filepath}), cannot assign: {filepath}')
				fasta_filepath = filepath

			# Assign the failed file
			if filepath.endswith('.json'):
				if failed_filepath: raise Exception(f'Failed file already assigned ({failed_filepath}), cannot assign: {filepath}')
				failed_filepath = filepath

		# Confirm the insert is possible
		if not blast_filepath or not fasta_filepath: raise Exception(f'Both a BLAST and FASTA file are required to operate')

		# Assign the table
		sql_table_assign = config_data[schema]

		# Assign the sequencing columns
		sequence_std_cols = ['sequence_id', 'seq_len', 'seq_percent_ident', 
							 'seq_align_len', 'sequence', 'reads', 'sample_id', 
							 'bold_id', 'species', 'sequence_status']
		
		# Assign the sequencing header data				 
		sequence_label_dict = {'Query ID': 'sequence_id', 
							   'Percent Identity': 'seq_percent_ident', 
							   'Alignment Length': 'seq_align_len',
							   'Query Length': 'seq_len'}

		# Index the sequence file
		sequence_index = SeqIO.index(fasta_filepath, 'fasta')

		# Read in the file as a pandas dataframe, prep for database
		input_dataframe = pd.read_csv(blast_filepath, dtype = str, sep = '\t')

		# Clean up the BLAST dataframe
		input_dataframe['sequence'] = input_dataframe['Query ID'].apply(append_seq, seq_index = sequence_index)
		input_dataframe[['Query ID', 'reads']] = input_dataframe['Query ID'].str.split(';size=', expand = True) 
		input_dataframe['sample_id'] = input_dataframe['Query ID'].str.split('_', expand = True)[0]
		input_dataframe[['bold_id', 'species']] = input_dataframe['Subject ID'].str.split('|', expand = True)[[0, 1]]
		input_dataframe['species'] = input_dataframe['species'].str.replace('_', ' ')
		input_dataframe['bold_id'] = input_dataframe['bold_id'].str.replace('_', ' ')
		input_dataframe['sequence_status'] = 'Species Identified'
		input_dataframe = input_dataframe.rename(columns = sequence_label_dict)
		input_dataframe = input_dataframe.replace(np.nan, None)

		# Remove the non standard columns
		sequence_non_std_cols = list(set(input_dataframe.columns) - set(sequence_std_cols))
		input_dataframe = input_dataframe.drop(columns =  sequence_non_std_cols)

		sql_insert = SQLInsert.fromConfig(config_data, sql_connection)
		sql_insert.addTableToInsert(sql_table_assign)
		sql_insert.addDataFrameValues(input_dataframe)
		sql_insert.insert()

		# Open the failed file
		with open(failed_filepath) as failed_file:

			# Assign the sequencing header data				 
			failed_label_dict = {'Query ID': 'sequence_id', 
								 'Status': 'sequence_status',
								 'Species': 'ambiguous_hits',
								 'Bins': 'bold_bins'}

			# Load the JSON into a dataframe
			failed_data = json.load(failed_file)
			failed_dataframe = pd.DataFrame(failed_data)
			
			# Join the lists, if found
			failed_dataframe['Species'] = failed_dataframe['Species'].apply(join_list)
			failed_dataframe['Bins'] = failed_dataframe['Bins'].apply(join_list)

			# Clean up the dataframe
			failed_dataframe['sequence'] = failed_dataframe['Query ID'].apply(append_seq, seq_index = sequence_index)
			failed_dataframe[['Query ID', 'reads']] = failed_dataframe['Query ID'].str.split(';size=', expand = True)
			failed_dataframe['sample_id'] = failed_dataframe['Query ID'].str.split('_', expand = True)[0] 
			failed_dataframe = failed_dataframe.rename(columns = failed_label_dict)
			failed_dataframe = failed_dataframe.replace(np.nan, None)
			
			sql_insert = SQLInsert.fromConfig(config_data, sql_connection)
			sql_insert.addTableToInsert(sql_table_assign)
			sql_insert.addDataFrameValues(failed_dataframe)
			sql_insert.insert()

		sequence_index.close()

	except:
		sql_connection.rollback()
		raise

	else:
		sql_connection.commit()

def insertStorageFileUsingConfig (config_file, filepath, schema = 'storage', plates_schema = 'plates', boxes_schema = 'boxes'):

	# Open the config and start the SQL session
	config_data = ConfigDB.readConfig(config_file)
	sql_connection = startSessionFromConfig(config_data)

	try:
		# Read in the file as a pandas dataframe, prep for database
		input_dataframe = pd.read_csv(filepath, dtype = str, sep = '\t')

		# Create the storage dataframe and prepare
		storage_dataframe = input_dataframe.copy()
		storage_dataframe = prepDataFrameUsingConfig(config_data, schema, storage_dataframe)
		storage_dataframe = storage_dataframe.replace(np.nan, None)

		# Insert the plates, if not already inserted
		plate_dataframe = storage_dataframe[['plate']].copy()
		plate_dataframe = plate_dataframe.drop_duplicates()
		plate_insert = SQLInsert.fromConfig(config_data, sql_connection)
		plate_insert.addTableToInsert(config_data['plates'])
		plate_insert.addDataFrameValues(plate_dataframe)
		plate_insert.addIgnore('plate')
		plate_insert.insertIgnore()

		# Confirm the input has a Box column
		if 'Box' in input_dataframe.columns:

			# Insert the boxes, if not already inserted
			box_dataframe = input_dataframe[['Box']].copy()
			box_dataframe = prepDataFrameUsingConfig(config_data, 'boxes', box_dataframe)
			box_dataframe = box_dataframe.drop_duplicates()
			box_dataframe = box_dataframe.replace(np.nan, None)
			box_insert = SQLInsert.fromConfig(config_data, sql_connection)
			box_insert.addTableToInsert(config_data['boxes'])
			box_insert.addDataFrameValues(box_dataframe)
			box_insert.addIgnore('box')
			box_insert.insertIgnore()
		
			# Create a dataframe with just the plates and boxes
			plate_update_dataframe = input_dataframe[['Plate', 'Box']].copy()
			plate_update_dataframe = plate_update_dataframe.rename(columns = {'Plate':'plate', 'Box':'box'})
			plate_update_dataframe = plate_update_dataframe.drop_duplicates()
			plate_update_dataframe = plate_update_dataframe.replace('', None)
			plate_update_dataframe = plate_update_dataframe.dropna(axis = 'index')

			# Check for possible assignment errors
			if plate_update_dataframe['plate'].duplicated().any(): raise Exception('Plate found in more than one box')

			# Convert the dataframe into a list of dicts
			for update_dict in plate_update_dataframe.to_dict('records'):

				# Create the where and value dicts
				where_dict = {'plate': update_dict['plate']}
				value_dict = {key:value for key, value in update_dict.items() if key != 'plate'}

				# Confirm there are values to add
				if not value_dict: continue 

				plate_update = SQLUpdate.fromConfig(config_data, sql_connection)
				plate_update.addTableToUpdate(config_data['plates'])
				plate_update.addDictValues(value_dict)
				plate_update.addDictWhere(where_dict, dict_type = 'Column', include = True, cmp_type = 'EQ')
				plate_update.update()

			# Insert a boxes dataframe, for updating if possible
			box_update_dataframe = input_dataframe.copy()
			box_update_dataframe = prepDataFrameUsingConfig(config_data, 'boxes', box_update_dataframe)
			box_update_dataframe = box_dataframe.drop_duplicates()
			box_update_dataframe = box_update_dataframe.replace('', None)
			box_update_dataframe = box_update_dataframe.dropna(axis = 'index')

			# Check for possible assignment errors
			if box_update_dataframe['box'].duplicated().any(): raise Exception('Box found in multiple locations')

			# Convert the dataframe into a list of dicts
			for update_dict in box_update_dataframe.to_dict('records'):

				# Create the where and value dicts
				where_dict = {'box': update_dict['box']}
				value_dict = {key:value for key, value in update_dict.items() if key != 'box'}

				# Confirm there are values to add
				if not value_dict: continue 

				box_update = SQLUpdate.fromConfig(config_data, sql_connection)
				box_update.addTableToUpdate(config_data['boxes'])
				box_update.addDictValues(value_dict)
				box_update.addDictWhere(where_dict, dict_type = 'Column', include = True, cmp_type = 'EQ')
				box_update.update()

		storage_insert = SQLInsert.fromConfig(config_data, sql_connection)
		storage_insert.addTableToInsert(config_data[schema])
		storage_insert.addDataFrameValues(storage_dataframe)
		storage_insert.insert()

	except:
		sql_connection.rollback()
		raise

	else:
		sql_connection.commit()


import os
import sys
import csv
import json
import logging

from Bio import SeqIO

from kocher_tools.database import insertValues, updateValues, confirmValue
from kocher_tools.storage import convertPlateWell

def assignStorageIDs (cursor, table, id_key, blast_filename, failed_filename):

	# Create dict to store ID assignsments
	id_assignments = {}

	# Open the blast file
	with open(blast_filename) as blast_file:

		# Read the data file
		blast_reader = csv.reader(blast_file, delimiter = '\t')

		if sys.version_info[0] == 3:

			# Save the header
			header = next(blast_reader)

		else:
		
			# Save the header
			header = blast_reader.next()

		# Assign the query ID column
		query_id_index = header.index('Query ID')

		# Loop the rows of data
		for row in blast_reader:

			# Assign query information that needs formatting
			query = row[query_id_index]
			query_seqID = query.split(';')[0]
			query_plate = query.split('-')[0]
			query_well = query.split('-')[1].rsplit('_', 1)[0]

			# Assign the unique ID using the well and plate
			id_assignments[query_seqID] = convertPlateWell(cursor, table, id_key, query_plate, query_well)

	# Check if a failed filename was assigned
	if failed_filename:

		# Open the failed file
		with open(failed_filename) as failed_file:

			# Read in the JSON data
			failed_data = json.load(failed_file)

			# Loop the JSON data, sample by sample
			for failed_sample_dict in failed_data:
				
				# Get the name of the failed sample
				failed_sample = failed_sample_dict['Query ID']

				# Assign information that needs formatting
				failed_sample_seqID = failed_sample.split(';')[0]
				failed_sample_plate = failed_sample.split('-')[0]
				failed_sample_well = failed_sample.split('-')[1].rsplit('_', 1)[0]


				# Assign the unique ID using the well and plate
				id_assignments[failed_sample_seqID] = convertPlateWell(cursor, table, id_key, failed_sample_plate, failed_sample_well)

	# Update log
	logging.info('ID assignment using Plate/Well successful')

	return id_assignments

def readSeqFiles (blast_filename, sequence_index, failed_filename, id_assignment_dict, id_key):

	# Open the blast file
	with open(blast_filename) as blast_file:

		# Read the data file
		blast_reader = csv.reader(blast_file, delimiter = '\t')

		if sys.version_info[0] == 3:

			# Save the header
			header = next(blast_reader)

		else:
		
			# Save the header
			header = blast_reader.next()

		# Assign the query ID column
		query_id_index = header.index('Query ID')

		# Assign the query length column
		query_len_index = header.index('Query Length')

		# Assign the percent identity column
		pct_ident_index = header.index('Percent Identity')
		
		# Assign the alignment length column
		aln_len_index = header.index('Alignment Length')

		# Assign the subject ID column
		subject_id_index = header.index('Subject ID')

		# Loop the rows of data
		for blast_line in blast_reader:

			# Assign query information that needs formatting
			query = blast_line[query_id_index]
			query_seqID = query.split(';')[0]
			query_abundance = query.split('size=')[1]

			# Obtain sequence from blast index 
			query_seq = sequence_index[query].format('fasta')

			# Convert the string into a blob
			query_seq_blob = sqlite3.Binary(query_seq.encode())

			# Assign the ID key value using the well and plate
			id_key_value = id_assignment_dict[query_seqID]

			# Assign the subject information that needs formatting
			subject = blast_line[subject_id_index]
			subject_ident = subject.split('|')[0]
			subject_species = subject.split('|')[1].replace('_', ' ')

			# Assign the header
			header = [id_key, 'Sequence ID', 'Species', 'Reads', 'BOLD Identifier', 'Percent Identity', 'Alignment Length', 'Sequence Length', 'Sequence', 'Status']

			# Assign the row values
			row = [id_key_value, query_seqID, subject_species, query_abundance, subject_ident, blast_line[pct_ident_index], blast_line[aln_len_index], blast_line[query_len_index], query_seq_blob, 'Species Identified']

			# Return the header and row
			yield header, row

	# Check if the failed filename was defined
	if failed_filename:

		# Open the failed file
		with open(failed_filename) as failed_file:

			# Read in the JSON data
			failed_data = json.load(failed_file)

			# Loop the JSON data, sample by sample
			for failed_sample_dict in failed_data:
				
				# Get the name of the failed sample
				failed_sample = failed_sample_dict['Query ID']

				# Assign information that needs formatting
				failed_sample_seqID = failed_sample.split(';')[0]
				failed_sample_abundance = failed_sample.split('size=')[1]

				# Obtain sequence from blast index 
				failed_sample_seq = sequence_index[failed_sample].format('fasta')

				# Convert the string into a blob
				failed_sample_seq_blob = sqlite3.Binary(failed_sample_seq.encode())

				# Assign the ID key value using the well and plate
				id_key_value = id_assignment_dict[failed_sample_seqID]

				# Assign the header
				header = [id_key, 'Sequence ID', 'Reads', 'Sequence', 'Status']

				# Assign the row values
				row = [id_key_value, failed_sample_seqID, failed_sample_abundance, failed_sample_seq_blob, failed_sample_dict['Status']]

				# Check if the failed status is Ambiguous Hits
				if failed_sample_dict['Status'] == 'Ambiguous Hits':

					# Extend the header 
					header.extend(['Ambiguous Hits', 'BOLD Bins'])

					# Create a string of the joined species
					species_str = ', '.join(failed_sample_dict['Species'])

					# Create a string of the joined bins
					bins_str = ', '.join(failed_sample_dict['Bins'])

					# Extend the row data
					row.extend([species_str, bins_str])

				# Check if the failed status has no hits
				else:

					# Extend the header 
					header.extend(['Ambiguous Hits', 'BOLD Bins'])

					# Extend the row data
					row.extend([None, None])

				# Return the header and row
				yield header, row

def addSeqFilesToDatabase (cursor, table, blast_filename, sequence_filename, failed_filename, storage_table, storage_key):

	# Create string for file output
	file_str = '%s, %s' % (blast_filename, sequence_filename)

	# Check if a failed file was specified
	if failed_filename:

		# Update the string
		file_str += ', %s' % failed_filename

	# Update log
	logging.info('Uploading barcode files (%s) to database' % file_str)

	# Assign the Unique IDs for each sample using the storage table
	id_assignment_dict = assignStorageIDs(cursor, storage_table, storage_key, blast_filename, failed_filename)

	# Index the sequence file
	sequence_index = SeqIO.index(sequence_filename, 'fasta')

	# Loop the loc file by line
	for header, seq_data in readSeqFiles(blast_filename, sequence_index, failed_filename, id_assignment_dict, storage_key):

		# Insert the sequence and speices into the database
		insertValues(cursor, table, header, seq_data)

	sequence_index.close()

	# Update log
	logging.info('Upload successful')

def updateSeqFilesToDatabase (cursor, table, select_key, blast_filename, sequence_filename, failed_filename, storage_table, storage_key):

	# Create string for file output
	file_str = '%s, %s' % (blast_filename, sequence_filename)

	# Check if a failed file was specified
	if failed_filename:

		# Update the string
		file_str += ', %s' % failed_filename

	# Update log
	logging.info('Uploading barcode files (%s) to database' % file_str)

	# Assign the Unique IDs for each sample using the storage table
	id_assignment_dict = assignStorageIDs(cursor, storage_table, storage_key, blast_filename, failed_filename)

	# Index the sequence file
	sequence_index = SeqIO.index(sequence_filename, 'fasta')

	# Loop the seq file by line
	for header, seq_data in readSeqFiles(blast_filename, sequence_index, failed_filename, id_assignment_dict, storage_key):

		# Check if the selection key isn't among the headers
		if select_key not in header:
			raise Exception('Selection key (%s) not found. Please check the input file' % select_key)

		# Create an empty string to store the select_value
		select_value = ''

		# Create an empty set dict
		seq_set_dict = {}

		# Create an empty selection dict
		seq_select_dict = {}
		seq_select_dict['IN'] = {}

		# Loop the header and seq data
		for header_column, seq_value in zip(header, seq_data):

			# Check if the current column is the selection key
			if header_column == select_key:

				# Update the select value
				select_value = seq_value

				# Populate the selection dict
				seq_select_dict['IN'][header_column] = [seq_value]

			else:

				# Populate the selection dict
				seq_set_dict[header_column] = seq_value

		# Check that selected value is present
		if not confirmValue(cursor, table, select_key, select_value):

			# If not, log warning
			logging.warning('Entry (%s) not found, unable to update. Please check input' % select_value)

		# Update the values for the selected value
		updateValues(cursor, table, seq_select_dict, seq_set_dict)

	sequence_index.close()

	# Update log
	logging.info('Upload successful')
